!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Original Module written by Julien Lesgourgues for using weak lensing data in CosmoMC        !
! Provides simple routines for inferring angular correlation functions from C_l's             !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This modified version was written by Jason Dossett.                     !
! Adapted to match equations in Schrabback et al [arXiv:0911.0053],       !                                                         !
! Last Modified 02/23/2011                                                !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module AngularS

  implicit none

  integer, parameter :: dl = KIND(1.d0)
  real(dl), parameter :: pi = 3.1415926535897932384626433832795_dl
  real(dl), parameter :: dbessel = 0.05_dl 
  real(dl), parameter :: spl_large=1.e40_dl
  real(dl), dimension(:), allocatable :: Bess0, ddBess0
  real(dl), dimension(:), allocatable :: Bess4, ddBess4

contains

  subroutine Cl_to_Ctheta(ll,C,ddC,theta,dtheta,pownorm,c1,c2,nlmax,nthetamax,nthetastep,nbinmax,xstop)

    integer, intent(in) :: nlmax,nthetamax,nbinmax,nthetastep
    real, intent(in) :: ll(nlmax),C(nlmax,nbinmax),ddC(nlmax,nbinmax)
    real(dl), intent(in) :: theta(nthetamax),pownorm(nthetamax),dtheta(nthetamax)
    real, intent(in) :: xstop
    real c1(nthetamax,nbinmax),c2(nthetamax,nbinmax)
    

    real, parameter :: dxmax=0.2  
    ! Step of integration such that J(x) cannot
    ! change to more than J(x+dxmax)
    ! 0.2 sufficient, no need to decrease

    ! for splint of Bessels
    real x,b0,b_lo,a0,fac1,fac2,Bessel0,Bessel4
    integer ntheta
    ! for splint of Cls
    integer k,khi,klo
    real a,b,h
    real Cl(nbinmax)
    ! for integration
    integer nl
    real lll,l_previous,lmax
    real integrand1(nbinmax),integrand1_previous(nbinmax)
    real integrand2(nbinmax),integrand2_previous(nbinmax)
    integer l_increment ! 0 for logarithmic, 1 for linear
    
    ! For Weighting integrals of the cross-correlation intergrals as described in [arXiv:0911.0053] and []  
    real c1t(4,nbinmax),c2t(4,nbinmax)
    real(dl)theta2,radtomin
    integer j,ntheta2
    ! loop over theta
    radtomin = 60._dl*180._dl/pi    
    do ntheta=1,nthetamax
      j=1
      c1(ntheta,:)=0.
      c2(ntheta,:)=0.
      do ntheta2=0,nthetastep-1
        theta2 = theta(ntheta)+ntheta2*dtheta(ntheta)
        ! maximum l value in the integral is the smallest of: 
        ! * the max l value at which C_l have been computed
        ! * or the value such that l*theta=xstop
        lmax=min(ll(nlmax),xstop/theta2)

        ! initialize the sum
        c1t(j,:)=0.
        c2t(j,:)=0.

        ! Limit l-->0 : the integrand vanishes (du to factor l)
        l_previous=0.
        integrand1_previous(:)=0.
        integrand2_previous(:)=0.

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! First non-zero l value in the integral !
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        nl=1
        lll=ll(nl)
        l_increment=0 ! start trying logarithmic increment

        do while (lll <= lmax)

          x=lll*theta2

          !Interpolate the Bessel functions
          b0 = x/dbessel
          b_lo = int(b0) +1 
          a0=  b_lo - b0                
          b0=  1._dl - a0 
          fac1 = a0*b0*dbessel**2/6
          fac2 = fac1*(a0-2)
          fac1 = fac1*(b0-2)
          Bessel0 = a0*Bess0(b_lo)+ b0*Bess0(b_lo+1) +fac1*ddBess0(b_lo) &
          +fac2*ddBess0(b_lo+1)
          Bessel4 = a0*Bess4(b_lo)+ b0*Bess4(b_lo+1) +fac1*ddBess4(b_lo) &
          +fac2*ddBess4(b_lo+1)

          !Interpolate C_l
          klo=1
          khi=nlmax
          do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(ll(k).gt.lll) then
              khi=k
            else
              klo=k
            endif
          end do
          h=ll(khi)-ll(klo)
          
          if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of C(l,nbin)'
            stop
          end if
          
          a=(ll(khi)-lll)/h
          b=(lll-ll(klo))/h
          Cl(:)=a*C(klo,:)+b*C(khi,:)+((a**3-a)*ddC(klo,:)+(b**3-b)*ddC(khi,:))*(h**2)/6.

          ! find integrand
          ! eq. 14 in Schrabback et al [arXiv:0911.0053].
          integrand1(:)=Cl(:)*(Bessel0)*lll/pi/2.
          integrand2(:)=Cl(:)*(Bessel4)*lll/pi/2.
          !          write(9,'(8E14.5)')lll,Cl(:)*lll,(Bessel0+Bessel4),integrand1(:)

          ! increment the sum
          c1t(j,:)=c1t(j,:)&
          +(0.5*(integrand1_previous(:)+integrand1(:))*(lll-l_previous))&
          *radtomin*theta2*(0.0004664+radtomin*theta2*(0.0044118-8.90878e-5*radtomin*theta2))
          c2t(j,:)=c2t(j,:)&
          +(0.5*(integrand2_previous(:)+integrand2(:))*(lll-l_previous))&
          *radtomin*theta2*(0.0004664+radtomin*theta2*(0.0044118-8.90878e-5*radtomin*theta2))

          !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          ! Next l value in the integral !
          !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          l_previous=lll

          if (l_increment == 0) then
            if (nl+1 <= nlmax) then ! test whether integral not already finished
              if ((ll(nl+1)-ll(nl))*theta2 <= dxmax) then
                nl=nl+1
                lll=ll(nl)   ! continue with logarithmic increment
              else
                l_increment=0 !switch to linear increment
                lll=l_previous+dxmax/theta2 ! linear increment
              end if
            else
              lll=lmax+1. ! integral already finished. This will stop the loops.
            end if
          else
            lll=l_previous+dxmax/theta2 ! linear increment
          end if

          integrand1_previous(:)=integrand1(:)
          integrand2_previous(:)=integrand2(:)

        end do
        !Simpsons 3/8 rule for weighting integral.     
        if(j==4)then
          c1(ntheta,:)= c1(ntheta,:)&
          +(3./8.*(c1t(1,:)+c1t(4,:))+9./8.*(c1t(2,:)+c1t(3,:)))*dtheta(ntheta)/pownorm(ntheta)
          c2(ntheta,:)= c2(ntheta,:)&
          +(3./8.*(c2t(1,:)+c2t(4,:))+9./8.*(c2t(2,:)+c2t(3,:)))*dtheta(ntheta)/pownorm(ntheta)

          c1t(1,:)=c1t(4,:)
          c2t(1,:)=c2t(4,:)
          j=2
        else
          j=j+1
        end if
      end do
    end do

  end subroutine Cl_to_Ctheta

  subroutine GetBessels(MaxArg)
    real(dl), intent(in):: MaxArg
    integer i
    real(dl), allocatable, dimension(:) :: x
    integer max_bes_ix
    integer, save :: last_max = 0

    max_bes_ix = nint(MaxArg / dbessel) + 3
    if (max_bes_ix > last_max) then
       last_max = max_bes_ix
       if (allocated(Bess0)) then
          deallocate(Bess0,ddBess0)
          deallocate(Bess4,ddBess4)
       end if
       allocate(Bess0(max_bes_ix),ddBess0(max_bes_ix))
       allocate(Bess4(max_bes_ix),ddBess4(max_bes_ix))

       allocate(x(max_bes_ix))
       Bess0(1)=1
       Bess4(1)=0
       x(1)=0
       do i=2, max_bes_ix
          x(i) = (i-1)*dbessel
          Bess0(i) = Bessj0(x(i)) 
          Bess4(i) = Bessj(4,x(i))
          if(Bess4(i) .NE. Bess4(i))then
          write(*,*) 'error in calculation of bessels'
          stop
          end if
       end do
       
       call spline(x,Bess0,max_bes_ix,spl_large,spl_large,ddBess0)
       call spline(x,Bess4,max_bes_ix,spl_large,spl_large,ddBess4)

       deallocate(x)
    end if

  end subroutine GetBessels

  FUNCTION bessj0(x)
    real(dl) bessj0,x
    real(dl) ax,xx,z
    real(dl) p1,p2,p3,p4,p5,q1,q2,q3,q4,q5,r1,r2,r3,r4,r5,r6, &
         s1,s2,s3,s4,s5,s6,y
    SAVE p1,p2,p3,p4,p5,q1,q2,q3,q4,q5,r1,r2,r3,r4,r5,r6,s1,s2,s3,s4, &
         s5,s6
    DATA p1,p2,p3,p4,p5/1.d0,-.1098628627d-2,.2734510407d-4, &
         -.2073370639d-5,.2093887211d-6/, q1,q2,q3,q4,q5/-.1562499995d-1, &
         .1430488765d-3,-.6911147651d-5,.7621095161d-6,-.934945152d-7/
    DATA r1,r2,r3,r4,r5,r6/57568490574.d0,-13362590354.d0, &
         651619640.7d0,-11214424.18d0,77392.33017d0,-184.9052456d0/,s1,s2, &
         s3,s4,s5,s6/57568490411.d0,1029532985.d0,9494680.718d0, &
         59272.64853d0,267.8532712d0,1.d0/

    if(abs(x).lt.8.d0)then
       y=x**2
       bessj0=(r1+y*(r2+y*(r3+y*(r4+y*(r5+y*r6)))))/(s1+y*(s2+y*(s3+y* &
            (s4+y*(s5+y*s6)))))
    else
       ax=abs(x)
       z=8.d0/ax
       y=z**2
       xx=ax-.785398164d0
       bessj0=sqrt(.636619772d0/ax)*(cos(xx)*(p1+y*(p2+y*(p3+y*(p4+y* &
            p5))))-z*sin(xx)*(q1+y*(q2+y*(q3+y*(q4+y*q5)))))
    endif

  END FUNCTION bessj0

  FUNCTION bessj1(X)
    real(dl), intent(in) :: x
    real(dl) bessj1,ax,z,xx
    real(dl) Y,P1,P2,P3,P4,P5,Q1,Q2,Q3,Q4,Q5,R1,R2,R3,R4,R5, &
         R6,S1,S2,S3,S4,S5,S6
    DATA R1,R2,R3,R4,R5,R6/72362614232.D0,-7895059235.D0,242396853.1D0,&
         -2972611.439D0,15704.48260D0,-30.16036606D0/, &
         S1,S2,S3,S4,S5,S6/144725228442.D0,2300535178.D0, &
         18583304.74D0,99447.43394D0,376.9991397D0,1.D0/
    DATA P1,P2,P3,P4,P5/1.D0,.183105D-2,-.3516396496D-4,.2457520174D-5, & 
         -.240337019D-6/, Q1,Q2,Q3,Q4,Q5/.04687499995D0,-.2002690873D-3, &     
         .8449199096D-5,-.88228987D-6,.105787412D-6/
    IF(ABS(X).LT.8.)THEN
       Y=X**2
       bessj1=X*(R1+Y*(R2+Y*(R3+Y*(R4+Y*(R5+Y*R6))))) &
            /(S1+Y*(S2+Y*(S3+Y*(S4+Y*(S5+Y*S6)))))
    ELSE
       AX=ABS(X)
       Z=8.0d0/AX
       Y=Z**2
       XX=AX-2.356194491d0
       bessj1=SQRT(.636619772d0/AX)*(COS(XX)*(P1+Y*(P2+Y*(P3+Y*(P4+Y &
            *P5))))-Z*SIN(XX)*(Q1+Y*(Q2+Y*(Q3+Y*(Q4+Y*Q5))))) &
            *SIGN(1._dl,x)
    ENDIF

  END FUNCTION bessj1

  FUNCTION bessj(N,X)
    real(dl) bessj
    real(dl), intent(in) :: x
    integer, intent(in) :: n
    integer, parameter :: IACC = 40
    real(dl), parameter :: BIGNO=1.d10,BIGNI=1.d-10
    integer jsum,j,m
    real(dl) bj,bjm, bjp, tox, sum

    IF(N.LT.2)STOP 'bad argument N in bessj'

    TOX=2/X
    IF(X.GT.FLOAT(N))THEN
       BJM=bessj0(X)
       BJ=bessj1(X)
       DO J=1,N-1
          BJP=J*TOX*BJ-BJM
          BJM=BJ
          BJ=BJP
       END DO
       bessj=BJ
    ELSE
       M=2*((N+INT(SQRT(FLOAT(IACC*N))))/2)
       bessj=0.0d0
       JSUM=0
       SUM=0._dl
       BJP=0._dl
       BJ=1._dl
       DO J=M,1,-1
          BJM=J*TOX*BJ-BJP
          BJP=BJ
          BJ=BJM
          IF(ABS(BJ).GT.BIGNO)THEN
             BJ=BJ*BIGNI
             BJP=BJP*BIGNI
             bessj=bessj*BIGNI
             SUM=SUM*BIGNI
          ENDIF
          IF(JSUM.NE.0)SUM=SUM+BJ
          JSUM=1-JSUM
          IF(J.EQ.N)bessj=BJP
       end do
       SUM=2.*SUM-BJ
       bessj=bessj/SUM
    ENDIF

  END FUNCTION bessj

  SUBROUTINE spline(x,y,n,yp1,ypn,y2)
    implicit none
    INTEGER, intent(in) :: n
    real(dl), intent(in) :: x(n), y(n), yp1, ypn
    real(dl), intent(out) :: y2(n)
    INTEGER i,k
    real(dl) p,qn,sig,un
    real(dl), dimension(:), allocatable :: u


    Allocate(u(1:n))
    if (yp1.gt..99d30) then
       y2(1)=0._dl
       u(1)=0._dl
    else
       y2(1)=-0.5d0
       u(1)=(3._dl/(x(2)-x(1)))*((y(2)-y(1))/(x(2)-x(1))-yp1)
    endif

    do i=2,n-1
       sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
       p=sig*y2(i-1)+2._dl 

       y2(i)=(sig-1._dl)/p

       u(i)=(6._dl*((y(i+1)-y(i))/(x(i+ &
            1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig* &
            u(i-1))/p
    end do
    if (ypn.gt..99d30) then
       qn=0._dl
       un=0._dl
    else
       qn=0.5d0
       un=(3._dl/(x(n)-x(n-1)))*(ypn-(y(n)-y(n-1))/(x(n)-x(n-1)))
    endif
    y2(n)=(un-qn*u(n-1))/(qn*y2(n-1)+1._dl)
    do k=n-1,1,-1
       y2(k)=y2(k)*y2(k+1)+u(k)
    end do

    Deallocate(u)

  END SUBROUTINE spline

end module AngularS

